/*
 * BDMA emulation
 *
 */

#include "sim.h"

/*
 * bdma_loadimage(filename)
 *
 * Loads ROM data from filename into the BDMA emulation buffer.
 *
 */
int bdma_loadimage(char *file)
{
  int fd,i;
  
  if ((fd = open(file, O_RDONLY)) < 0)
    {
      perror("bdma_loadimage: open");
      return fd;
    }
  
  memset(curdsp.bdmasrc, 0, BDMASRC_SIZE); /* zero it out first */
  errno = 0;
  for (i=0; (i < BDMASRC_SIZE); i++)
    {
      curdsp.bdmasrc[i] = nextbyte(fd);
      if ((curdsp.bdmasrc[i] == 0xff) && (errno!=0))
	fprintf(stderr, "Error!\n");
    }

  close(fd);
  return 0;
}

/*
 * bdma_dotransfer()
 *
 * Called whenever MMAPREG_BWCOUNT is set.  Actually transfers the
 * data from bdmasrc[] to dm[] using the values from the mmap registers.
 *
 */
int bdma_dotransfer(void)
{
  static int bdma_inprogress = 0;
  int external, internal, len;

  if (bdma_inprogress)
    return -1;  /* return if we're already here */
  
  switch (getmmapreg(MMAPREG_BDMACTL) & 0x3)
    {
    case 0x01: /* 16bit DM */
      /*
       * In 16bit mode, we're given BIAD and BWCOUNT in 16bit words and 
       * BEAD in bytes..
       */
      external = getmmapreg(MMAPREG_BEAD)/2;
      internal = getmmapreg(MMAPREG_BIAD);
      len = getmmapreg(MMAPREG_BWCOUNT);

      fprintf(stderr, "bdma_dotransfer: Starting 16bit DM BDMA transfer from 0x%04x into 0x%04x of %d bytes...\n", external*2, internal, len);
      
      bdma_inprogress = 1; /* lock the DMA channel */
      while (getmmapreg(MMAPREG_BWCOUNT))
	{
	  setdmword(internal, bdma_getdword(external));
	  internal++;
	  external++; 
	  
	  /* must update BWCOUNT in case someone is watching */
	  len--;
	  setmmapreg(MMAPREG_BWCOUNT, len); 
	}
      bdma_inprogress = 0; /* unlock DMA channel */

      break;
    case 0x02: /* 8bit DM MSB-aligned */
    case 0x03: /* 8bit DM LSB-aligned */
      fprintf(stderr, "bdmatransfer: unknown BTYPE.  failing\n");
      exit(0);
    case 0x00: /* 24bit PM */
      /* 
       * In 24bit mode, we're given the IAD in iwords, the EAD in bytes, and
       * the BWCOUNT in words.  Yes, that is fucked.  
       */
      external = getmmapreg(MMAPREG_BEAD)/3;
      internal = getmmapreg(MMAPREG_BIAD);
      len = getmmapreg(MMAPREG_BWCOUNT);
      
      fprintf(stderr, "bdma_dotransfer: Starting 24bit PM BDMA transfer from 0x%04x into 0x%04x of %d bytes...\n", external*3, internal, len);
      /* 
       * We'll do this the Right Way, in case we ever go multithreaded and
       * fully emulate how the DSP does this...
       */
      bdma_inprogress = 1; /* lock the DMA channel */
      while (getmmapreg(MMAPREG_BWCOUNT))
	{
	  pmsetiword(internal, bdma_getiword(external));
	  internal++;
	  external++; 
	  
	  /* must update BWCOUNT in case someone is watching */
	  len--;
	  setmmapreg(MMAPREG_BWCOUNT, len); 
	}
      bdma_inprogress = 0; /* unlock DMA channel */
    } /* switch */

  fprintf(stderr, "bdma_dotransfer: BDMA Trasfer Complete\n");

  if (getmmapreg(MMAPREG_BDMACTL) & 0x8)
    init_softreboot();
  
  irq_flag(8);
	  
  return 0;
}

unsigned int bdma_getiword(unsigned int waddr)
{
  unsigned int baddr;
  unsigned int iword;
  baddr = waddr*3; /* convert 24bit word address into 8bit byte address */
  
  iword = glue24(curdsp.bdmasrc[baddr],curdsp.bdmasrc[baddr+1],curdsp.bdmasrc[baddr+2]);
  return iword;
}

unsigned int bdma_getdword(unsigned int waddr)
{
  unsigned int baddr;
  unsigned int dword;
  baddr = waddr*2; /* convert 16bit word address into 8bit byte address */
  
  dword = glue16(curdsp.bdmasrc[baddr],curdsp.bdmasrc[baddr+1]);
  return dword;
}
